from typing import Optional

from popupdict.config import *
from .abstract import AbstractQueryClient
from ..result import QueryResult


# 有道词典网页版
class YoudaoWebQueryClient(AbstractQueryClient):
    id = 'youdao-web'
    config_class = QueryClientConfiguration

    def query(self, word: str) -> Optional[QueryResult]:
        raise Exception('Client not implemented!')



import requests
import random
import hashlib
import json
from typing import Optional

from popupdict.config import *
from popupdict.util import logger
from .abstract import AbstractQueryClient
from ..result import QueryResult


# 有道智云的自然语言翻译服务
class YoudaoZhiyunQueryClient(AbstractQueryClient):
    id = 'youdao-zhiyun'
    config_class = YoudaoZhiyunConfiguration

    API = 'https://openapi.youdao.com/api'
    DICT_LINK = 'https://dict.youdao.com/w/{}/{}/'

    def __init__(self, config: YoudaoZhiyunConfiguration):
        super().__init__(config)
        self.config = config

    def query(self, text: str) -> Optional[QueryResult]:
        try:
            res = requests.get(self.API, params=self._params(text), timeout=self.config.request_timeout)
            if not res.ok:
                logger.error('[%s] Request failed for query=%s: status_code=%s, reason=%s, content=%s',
                             self.id, repr(text), res.status_code, res.reason, repr(res.text))
                return

            try:
                result = res.json()
            except json.decoder.JSONDecodeError:
                logger.error('[%s] Invalid JSON response for query=%s: ', self.id, repr(text), repr(res.text))
                return

            if result['errorCode'] != '0':
                logger.error('[%s] Query failed for query=%s: errorCode=%s',
                             self.id, repr(text), str(result['errorCode']))
                return

            logger.debug('[%s] Request success for query=%s: %s', self.id, repr(text), result)

            language = result['l'].replace('2zh-CHS', '').lower()
            translation = result['translation'] != [result['query']] and '; '.join(result['translation']) or None
            basic = result.get('basic')
            phrases = result.get('web') or None

            # 检查有无有效翻译结果。translation 和 explanations 至少要有其一
            # 特例：
            #   * 无 translation, 有 explanations: alt, github
            #   * 有 translation, 无 explanations: 除英语外的其它语种
            if not translation and (not basic or not basic.get('explains')):
                logger.debug('[%s] No translation for query=%s', self.id, repr(text))
                return

            if phrases is not None:
                for phrase in phrases:
                    phrase['value'] = '; '.join(phrase['value'])
                    phrase['dict_link'] = __class__.dict_link(language, phrase['key'])

            # 小语种查询结果没有 'query'
            query_result = QueryResult(result.get('query', text), translation,
                                       dict_link=__class__.dict_link(language, result.get('query', text)),
                                       phonetic=basic and (
                                           basic.get('us-phonetic') or
                                           basic.get('uk-phonetic') or
                                           basic.get('phonetic')),
                                       explanations=basic and basic.get('explains'),
                                       phrases=phrases
                                       )
            return query_result
        except requests.exceptions.RequestException:
            logger.exception('[%s] Request failed for query=%s', self.id, repr(text))

    @staticmethod
    def dict_link(lang, query):
        return __class__.DICT_LINK.format(lang, __class__.escape_url_path(query))

    # 构造请求参数
    def _params(self, text: str):
        # http://ai.youdao.com/docs/doc-trans-api.s#p02
        params = {
            'q': text,
            # 设置为 auto 会导致有时识别错误
            'from': 'EN',
            'to': 'zh-CHS',
            'appKey': self.config.app_id,
            'salt': random.randint(1, 10000),
        }
        s = self.config.app_id + text + str(params['salt']) + self.config.app_secret
        params['sign'] = hashlib.md5(s.encode()).hexdigest()
        return params



from abc import ABC, abstractmethod
from typing import Optional

from popupdict.config import SpeechClientConfiguration


class AbstractSpeechClient(ABC):
    # 客户端 ID，需唯一，用于区分客户端（配置文件、缓存等）
    id = 'abstract'
    # 配置类
    config_class = SpeechClientConfiguration

    def __init__(self, config: SpeechClientConfiguration):
        config.validate()
        self.config = config

    @abstractmethod
    # save_path: 建议的发音文件保存路径
    # 返回发音路径
    def get(self, text: str, save_path: str) -> Optional[str]:
        pass



import time
from typing import Optional

from popupdict.gtk import *
from popupdict.query import QueryResult
from popupdict.util import Selection, logger
from .widgets import Widgets


# 弹窗
class Popup(Gtk.Window):
    WINDOW_WIDTH = 400  # type: int

    def __init__(self, popup_timeout: float = None, auto_play: bool = True):
        super().__init__(type=Gtk.WindowType.POPUP)

        # 弹窗显示时间。单位：秒
        self.popup_timeout = popup_timeout or 3  # type: float
        # 是否自动发音
        self.auto_play = auto_play

        self.set_border_width(10)
        self.set_default_size(__class__.WINDOW_WIDTH, -1)
        self.set_size_request(__class__.WINDOW_WIDTH, -1)

        # Initialize container, widgets
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, halign=Gtk.Align.START)
        self.widgets = Widgets(self.box, self.pronounce)
        self.add(self.box)

        # Time to hide to window. Timestamp with float point got by time.time()
        self.time_to_hide = 0  # type: float

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clipboard.connect("owner-change", __class__.selection_changed)

        self.current_query_result = None  # type: Optional[QueryResult]
        self.player = Gst.ElementFactory.make('playbin', 'player')  # type: Gst.Element
        if not self.player:
            raise Exception('Unable to create audio player')

        # 鼠标设备
        self.pointer_device = Gdk.Display.get_default().get_default_seat().get_pointer()  # type: Gdk.Device
        # 弹窗显示一段时间后，自动隐藏
        GLib.timeout_add(300, self.hide_window_if_timeout)

    # 渲染翻译结果
    def redraw(self, selection: Selection, query_result: QueryResult):
        if not query_result:
            logger.debug('Redraw canceled, since query_result is invalid: %s', repr(query_result))
            return
        # 若选中文本已经变化，不再显示旧的查询结果
        if selection.text != Selection.current.text:
            logger.debug('Redraw canceled, since selection has changed: original=%s, current=%s',
                         repr(selection.text), repr(Selection.current.text))
            return
        logger.debug('Redraw started for query=%s', repr(selection.text))
        self.current_query_result = query_result
        self.widgets.draw(query_result)
        minimum_height, natural_height = self.box.get_preferred_height()
        # Avoid warning: Allocating size to window without calling gtk_widget_get_preferred_width/height()
        # self.get_preferred_width()
        # self.get_preferred_height()
        self.resize(self.WINDOW_WIDTH, minimum_height)
        self.move_window(selection)
        self.show()
        self.time_to_hide = time.time() + self.popup_timeout
        if self.auto_play and query_result.speech_path:
            self.pronounce(query_result.speech_path)

    # 发音下载完成后，决定是否发音
    def on_speech_downloaded(self, query: str, path: str):
        logger.debug('on_speech_downloaded: query=%s, path=%s', repr(query), repr(path))
        if not self.auto_play:
            return
        # 若窗口未显示，或查询内容已改变，不发音
        if not self.is_visible() or not self.current_query_result or self.current_query_result.query != query:
            return
        # 避免重复发音
        if not self.current_query_result.speech_path:
            self.pronounce(path)
            self.current_query_result.speech_path = path

    # 发音
    def pronounce(self, path: Optional[str] = None):
        path = path or (self.current_query_result and self.current_query_result.speech_path)
        if not path:
            logger.error("Cannot pronounce: path=%s, current_query_result.speech_path=%s",
                         repr(path), repr(self.current_query_result and self.current_query_result.speech_path))
            return
        self.player.set_state(Gst.State.READY)
        self.player.set_property('uri', Gst.filename_to_uri(path))
        self.player.set_state(Gst.State.PLAYING)

    # 根据选中文本位置、弹窗大小、屏幕大小确定弹窗位置，保证弹窗显示完整
    def move_window(self, selection: Selection):
        target_x, target_y = selection.position
        monitor_size = selection.monitor.get_geometry()
        width, height = self.get_size()

        def try_move(x, y):
            if x >= 0 and y >= 0 and x + width <= monitor_size.width and y + height <= monitor_size.height:
                self.move(x, y)
                return True
            return False

        # 下
        if try_move(target_x, target_y + 30):
            return
        # 上
        if try_move(target_x, target_y - height - 30):
            return
        # 左
        if try_move(target_x - width - 40, target_y - height / 2):
            return
        # 右
        if try_move(target_x + 40, target_y - height / 2):
            return
        self.move(10, 10)

    # 检查鼠标是否在窗口
    def is_pointer_in_window(self) -> bool:
        _, pointer_x, pointer_y = self.pointer_device.get_position()
        x, y = self.get_position()
        width, height = self.get_size()
        return x < pointer_x < x + width and y < pointer_y < y + height

    # 弹窗显示一段时间后，自动隐藏
    def hide_window_if_timeout(self) -> bool:
        if self.is_visible() and time.time() >= self.time_to_hide:
            # 若鼠标在窗口中，延迟隐藏窗口
            if not self.is_pointer_in_window():
                self.hide()
        return True

    # 选中文本变化事件处理
    @staticmethod
    def selection_changed(clipboard: Gtk.Clipboard, event: Gdk.EventOwnerChange):
        text = clipboard.wait_for_text()
        Selection.update(text, event)



from .base import QueryConfiguration, QueryClientConfiguration
from .youdao_zhiyun import YoudaoZhiyunConfiguration

__all__ = [
    'QueryConfiguration',
    'QueryClientConfiguration',
    'YoudaoZhiyunConfiguration',
]



import signal
import os.path
import sys
import psutil
import argparse
from typing import Optional, Dict

from .gtk import *
from .config import Configuration, ConfigError
from .query import QueryAdapter
from .speech import SpeechAdapter
from .ui import Popup
from .daemon import QueryDaemon

PID_FILE = os.path.join(GLib.get_user_cache_dir(), 'popup-dict', 'popup-dict.pid')


def read_pid() -> Optional[int]:
    if os.path.exists(PID_FILE):
        with open(PID_FILE) as f:
            try:
                return int(f.read().strip())
            except ValueError:
                pass
    return None


def write_pid(pid: int):
    os.makedirs(os.path.dirname(PID_FILE), 0o755, True)
    with open(PID_FILE, 'w') as f:
        f.write(str(pid))


def delete_pid():
    if os.path.exists(PID_FILE):
        os.remove(PID_FILE)


def is_running():
    pid = read_pid()
    if pid and psutil.pid_exists(pid):
        process = psutil.Process(pid)
        if 'dict' in process.name():
            return True
    return False


def start(config_file: str = None, cmd_config: Optional[Dict] = None):
    config = Configuration(config_file, cmd_config)
    query_adapter = QueryAdapter(config)
    speech_adapter = config.speech.enabled and SpeechAdapter(config) or None

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    window = Popup(config.popup_timeout, config.speech.auto_play)
    window.show_all()
    window.hide()

    QueryDaemon(window, query_adapter, speech_adapter).start()

    Gtk.main()


def main():
    parser = argparse.ArgumentParser(description='划词翻译')
    parser.add_argument('--client', metavar='CLIENT_ID', help='查询客户端')
    parser.add_argument('--debug', default=None, action='store_true', help='开启调试模式')
    parser.add_argument('--no-debug', default=None, action='store_false', help='关闭调试模式', dest='debug')
    parser.add_argument('-c', '--config', metavar='CONFIG_FILE',
                        help='配置文件，默认 ~/.config/popup-dict/config.ini, 或 /etc/popup-dict/config.ini')
    args = parser.parse_args()

    if args.config and not os.path.exists(args.config):
        print("Config file {} does not exist!".format(repr(args.config)), file=sys.stderr)
        sys.exit(1)

    # 命令行配置项，用于覆盖配置文件中的配置
    cmd_config = {
        'global': {
            'debug': args.debug,
        },
        'query': {
            'client': args.client,
        },
    }

    # 未指定命令行参数时避免覆盖配置文件
    cmd_config['global'] = {k: v for k, v in cmd_config['global'].items() if v is not None}
    cmd_config['query'] = {k: v for k, v in cmd_config['query'].items() if v is not None}

    if is_running():
        print('An instance is already running!', file=sys.stderr)
        sys.exit(2)
    else:
        write_pid(os.getpid())
        try:
            start(args.config, cmd_config)
        except ConfigError as e:
            print(e.args[0], file=sys.stderr)
            sys.exit(1)


if __name__ == '__main__':
    main()



import time
import threading
from typing import Optional
from concurrent.futures import Future, ThreadPoolExecutor

from popupdict.gtk import *
from popupdict.ui import Popup
from popupdict.query import QueryAdapter, QueryResult
from popupdict.speech import SpeechAdapter
from popupdict.util import Selection


# 监视选中文本，如果变化，发起翻译请求，翻译完成后通知主线程渲染 UI
# 为何使用单独的线程：
#   1. 避免主线程繁忙/阻塞，影响响应速度
#   2. 为避免选中文本过程中频繁触发查询，需要延时查询（即选中内容不再变化时再查询）。使用线程更易处理
# 有些软件（如 Gnome Terminal）中释放鼠标后才会触发 selection change 事件，而有些软件（如 IntelliJ IDEA）在选中过程中会不停触发
class QueryDaemon(threading.Thread):
    # 查询延时（选中文本多长时间不再变化时查询）。单位：秒
    QUERY_DELAY = 0.1

    def __init__(self, popup: Popup,
                 query_adapter: QueryAdapter,
                 speech_adapter: Optional[SpeechAdapter]):
        super().__init__(daemon=True)
        self.popup = popup
        self.query_adapter = query_adapter
        self.speech_adapter = speech_adapter
        self.executor = ThreadPoolExecutor(20)

    def run(self):
        last_persist_cache_time = time.time()
        last_selection = None  # type: Optional[Selection]
        # 查询翻译任务
        query_task = None  # type: Optional[Future]
        # 下载发音任务
        speech_task = None  # type: Optional[Future]
        # 查询翻译、下载发音两个任务的完成顺序是不确定的。后者是次要功能，不应拖慢 UI 响应
        # 如果查询翻译完成时，下载发音还未完成，立即渲染 UI，不必等待。待发音下载完成后，由 UI 组件决定是否发音
        while True:
            # 翻译完成，通知主线程渲染
            if query_task and query_task.done():
                query_result = query_task.result()  # type: QueryResult
                query_task = None
                if query_result:
                    # 若发音已下载完成，优先使用
                    if speech_task:
                        if speech_task.done():
                            query_result.speech_path = speech_task.result()
                            speech_task = None
                        # 若翻译结果中已有发音，取消下载发音任务
                        elif query_result.speech_path:
                            speech_task.cancel()
                            speech_task = None
                    # 只在主线程中渲染 UI
                    GLib.idle_add(self.popup.redraw, last_selection, query_result)

            # 翻译结果渲染后，发音才下载好。
            if not query_task and speech_task and speech_task.done():
                speech_path = speech_task.result()
                speech_task = None
                if speech_path:
                    GLib.idle_add(self.popup.on_speech_downloaded, last_selection.text, speech_path)

            # 选中内容不再变化时，发起翻译请求。避免选中文本过程中频繁请求
            selection = Selection.current
            if selection and not selection.queried and (selection.time + __class__.QUERY_DELAY
                                                        <= time.time() <=
                                                        selection.time + self.popup.popup_timeout):
                # 取消未完成的查询请求
                # TODO 如果请求已经在执行，Future.cancel 并不能真正取消
                if query_task:
                    query_task.cancel()
                if speech_task:
                    speech_task.cancel()

                last_selection = selection
                query_task = self.executor.submit(self.query_adapter.query, selection.text)
                if self.speech_adapter:
                    speech_task = self.executor.submit(self.speech_adapter.get, selection.text)

                selection.queried = True

            # 定期持久化缓存到文件系统
            if time.time() - last_persist_cache_time > 5:
                self.query_adapter.persist_cache()
                if self.speech_adapter:
                    self.speech_adapter.remove_old_cache()
                last_persist_cache_time = time.time()

            time.sleep(0.05)



from typing import Optional
from configparser import SectionProxy

from .base import ConfigError, QueryClientConfiguration


class YoudaoZhiyunConfiguration(QueryClientConfiguration):
    def __init__(self, section: SectionProxy):
        super().__init__(section)
        self.app_id = section.get('app_id')  # type: Optional[str]
        self.app_secret = section.get('app_secret')  # type: Optional[str]

    def validate(self):
        super().validate()
        if not self.app_id:
            raise ConfigError('app_id is empty!')
        elif not self.app_secret:
            raise ConfigError('app_secret is empty!')



from typing import List, Type

from .abstract import AbstractQueryClient
from .fake import FakeQueryClient
from .youdao_web import YoudaoWebQueryClient
from .youdao_zhiyun import YoudaoZhiyunQueryClient

valid_clients: List[Type[AbstractQueryClient]] = [
    FakeQueryClient,
    YoudaoWebQueryClient,
    YoudaoZhiyunQueryClient,
]

__all__ = [
    'valid_clients',
    'AbstractQueryClient',
    'FakeQueryClient',
    'YoudaoWebQueryClient',
    'YoudaoZhiyunQueryClient',
]



from .exception import ConfigError
from .configuration import Configuration
from .query import *
from .speech import *

__all__ = [
    'ConfigError',
    'Configuration',
    'QueryClientConfiguration',
    'YoudaoZhiyunConfiguration',
    'SpeechConfiguration',
    'SpeechClientConfiguration',
]



from typing import Optional
import requests

from .abstract import AbstractSpeechClient
from popupdict.util import logger


class YoudaoSpeechClient(AbstractSpeechClient):
    id = 'youdao'

    def get(self, text: str, save_path: str) -> Optional[str]:
        params = {'le': 'en', 'audio': text}
        # noinspection PyBroadException
        try:
            res = requests.get('http://dict.youdao.com/dictvoice', params=params, timeout=self.config.request_timeout)
            if res.status_code != 200 or len(res.content) < 1000:
                return
            with open(save_path, 'wb') as f:
                f.write(res.content)

            return save_path

        except Exception:
            logger.exception('[speech %s] Request failed for text=%s', self.id, repr(text))
            return



import os.path
import logging
from configparser import ConfigParser, SectionProxy
from typing import Optional, List, Dict
import pkg_resources

from popupdict.gtk import *
from popupdict.util import logger, config_dir
from .exception import ConfigError
from .query import QueryConfiguration
from .speech import SpeechConfiguration


class Configuration:
    # 默认配置，始终加载
    DEFAULT_CONFIG = pkg_resources.resource_string(__package__, 'default.ini').decode()
    # 候选配置文件，按顺序检查，只加载第一个存在的文件
    CANDIDATE_CONFIG_FILES = [
        os.path.join(config_dir, 'config.ini'),
        '/etc/popup-dict/config.ini',
    ]  # type: List[str]

    def __init__(self, config_file: Optional[str] = None, extra_config: Optional[Dict] = None):
        if config_file:
            if not os.path.exists(config_file):
                raise ConfigError("Config file {} does not exist!".format(repr(config_file)))
        else:
            # 若未指定配置文件，从候选配置文件中加载第一个存在的文件
            for path in __class__.CANDIDATE_CONFIG_FILES:
                if os.path.exists(path):
                    config_file = path
                    break

        parser = self.get_parser(None, config_file, extra_config)

        # 全局配置
        try:
            global_section = parser['global']

            # 调试模式
            self.debug = global_section.get('debug') and global_section.getboolean('debug', False) or False
            logger.setLevel(self.debug and logging.DEBUG or logging.INFO)

            # 是否启用缓存
            self.cache = global_section.getboolean('cache')
            # 最大缓存条目数。仅针对单个查询客户端、发音客户端
            self.max_cache_items = global_section.getint('max_cache_items')

            # 弹窗显示时间
            self.popup_timeout = global_section.getfloat('popup_timeout')

            # Gtk Global Dark Theme
            # 不设置或设为空则使用系统默认设置
            prefer_dark_theme = global_section.get('prefer_dark_theme')
            if prefer_dark_theme is not None and prefer_dark_theme != '':
                self.prefer_dark_theme = global_section.getboolean('prefer_dark_theme')
                Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", self.prefer_dark_theme)
            else:
                self.prefer_dark_theme = None

        except KeyError as e:
            raise ConfigError('Missing global configuration: ' + e.args[0])

        # 查询配置
        self.query = QueryConfiguration(parser['query'])
        # 发音配置
        self.speech = SpeechConfiguration(parser['speech'])

        # 各查询客户端配置
        # 为避免 circular imports, 不能在此处实例化 ClientConfiguration 的子类
        parser = self.get_parser('query-client', config_file, extra_config)
        self.clients = {'default': parser['query-client']}  # type: Dict[str, SectionProxy]
        for section_name in parser.sections():
            if section_name.startswith('query:'):
                self.clients[section_name.replace('query:', '')] = parser[section_name]

        # 各发音客户端配置
        parser = self.get_parser('speech-client', config_file, extra_config)
        self.speech_clients = {'default': parser['speech-client']}  # type: Dict[str, SectionProxy]
        for section_name in parser.sections():
            if section_name.startswith('speech:'):
                self.speech_clients[section_name.replace('speech:', '')] = parser[section_name]

    def __repr__(self):
        return repr(__dict__)

    # 解析配置
    # ConfigParser 的 default_section 不能动态修改，因此需要不同的 default_section 时重新解析配置
    @staticmethod
    def get_parser(default_section: Optional[str],
                   config_file: Optional[str] = None,
                   extra_config: Optional[Dict] = None) -> ConfigParser:
        parser = ConfigParser(default_section=default_section)
        parser.read_string(__class__.DEFAULT_CONFIG)

        if config_file:
            parser.read(config_file)

        if extra_config:
            parser.read_dict(extra_config)

        return parser



import os.path
from typing import Optional
import tempfile
import time

from popupdict.config import Configuration, ConfigError
from popupdict.util import cache_dir, logger
from .client import *


class SpeechAdapter:
    def __init__(self, config: Configuration):
        for client_class in valid_speech_clients:
            if client_class.id == config.speech.client_id:
                config_section = config.speech_clients.get(client_class.id, config.speech_clients['default'])
                client_config = client_class.config_class(config_section)
                self.client = client_class(client_config)
                break
        if not self.client:
            raise ConfigError("Unknown speech client: {}".format(repr(config.speech.client_id)))

        self.cache_enabled = config.cache
        self.max_cache_items = config.max_cache_items
        if self.cache_enabled:
            self.cache_dir = os.path.join(cache_dir, 'speech', self.client.id)  # type: str
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir, 0o755)
        else:
            self.tmp_dir = tempfile.mkdtemp(prefix='popup-dict-speech-' + self.client.id + '.')  # type: str

    def get(self, text: str) -> Optional[str]:
        start_time = time.time()
        logger.debug('[speech %s] Get speech started for text=%s', self.client.id, repr(text))

        # noinspection PyBroadException
        try:
            # 检查缓存
            if self.cache_enabled:
                path = os.path.join(self.cache_dir, text)
                if os.path.exists(path) and os.path.getsize(path) > 0:
                    logger.debug('[speech %s] Cache hit for text=%s', self.client.id, repr(text))
                    return path
            else:
                path = os.path.join(self.tmp_dir, text)
                if os.path.exists(path):
                    return path

            path = self.client.get(text, path)
            logger.debug('[speech %s] Get speech finished for text=%s, path=%s', self.client.id, repr(text), repr(path))
            return path

        except Exception:
            logger.exception('[speech %s] Get speech failed for text=%s', self.client.id, repr(text))
        finally:
            time_spent = (time.time() - start_time) * 1000
            logger.debug('[speech %s] Time spent: %fms', self.client.id, time_spent)

    # 缓存条目数超出数量时，删除旧缓存。使用 LRU 原则
    def remove_old_cache(self):
        start_time = time.time()
        files = os.listdir(self.cache_dir)
        if len(files) <= self.max_cache_items:
            return
        files = [os.path.join(self.cache_dir, f) for f in files]
        pairs = [(os.path.getatime(f), f) for f in files]
        pairs.sort()
        for pair in pairs[1000:]:
            os.remove(pair[1])
        time_spent = (time.time() - start_time) * 1000
        logger.debug('Time spent for clearing old speech cache: %fms', time_spent)



import os.path

from popupdict.gtk import GLib

config_dir = os.path.join(GLib.get_user_config_dir(), 'popup-dict')

cache_dir = os.path.join(GLib.get_user_cache_dir(), 'popup-dict')
if not os.path.exists(cache_dir):
    os.makedirs(cache_dir, 0o755)



import time
from typing import Optional, Tuple

from popupdict.gtk import Gdk
from .selection_filter import SelectionFilter
from .logging import logger


# 选中文本状态
class Selection:
    # 当前/最近选中状态
    current = None  # type: Optional[Selection]

    def __init__(self, text: str, event: Gdk.EventOwnerChange):
        # 是否已翻译过
        self.queried = False
        # 选中文本
        self.text = text
        # 选中时间
        # event.selection_time 不知道是怎么计算的时间戳，无法与当前时间比较
        self.time = time.time()

        # 屏幕坐标
        # 获取到的是实时坐标，查询翻译后再获取可能会有较大偏差
        _, x, y, mods = event.window.get_pointer()
        self.position = (x, y)  # type: Tuple[int, int]

        # 显示器，用于获取屏幕大小以决定弹窗位置，保证显示完整
        self.monitor = event.window.get_display().get_monitor_at_window(event.window)  # type: Gdk.Monitor

    # 每次变化时，创建一个新的对象（避免修改原对象），以避免线程同步问题
    @staticmethod
    def update(text: str, event: Gdk.EventOwnerChange):
        logger.debug('Selection Changed: %s', repr(text))
        text = SelectionFilter.filter(text)
        logger.debug('Filtered as: %s', repr(text))
        if text:
            __class__.current = Selection(text, event)



from popupdict.gtk import *
from popupdict.query import QueryResult
from typing import Callable


# UI 元素
class Widgets:
    # 查询内容模板
    QUERY_TEMPLATE = '<a href="{}"><span size="xx-large" underline="none" color="#2980B9">{}</span></a>'
    QUERY_TEMPLATE_WITHOUT_LINK = '<span size="xx-large">{}</span>'
    # 词组短语列表项模板
    PHRASE_ITEM_TEMPLATE = '<a href="{}"><span underline="none" color="#2980B9">{}</span></a>: <span>{}</span>'
    PHRASE_ITEM_TEMPLATE_WITHOUT_LINK = '<span>{}</span>: <span>{}</span>'

    def __init__(self, container: Gtk.Box, pronounce_func: Callable):
        # 查询内容
        self.query = __class__.single_line_label()  # type: Gtk.Label
        container.pack_start(self.query, False, False, 0)

        # 音标
        # label 无法接收鼠标事件，因此使用 EventBox
        self.phonetic_event_box = Gtk.EventBox()
        self.phonetic = __class__.auto_wrap_label()
        self.phonetic_event_box.add(self.phonetic)
        self.phonetic_event_box.connect('button-press-event', lambda widget, event: pronounce_func())
        container.pack_start(self.phonetic_event_box, False, False, 0)

        # 简略释义
        self.translation = __class__.auto_wrap_label()  # type: Gtk.Label
        container.pack_start(self.translation, False, False, 20)

        # 基本翻译标题
        self.explanations_label = __class__.label(
            '<span size="large" weight="bold" alpha="95%">基本释义：</span>')  # type: Gtk.Label
        container.pack_start(self.explanations_label, False, False, 5)

        # 基本翻译列表
        self.explanations = __class__.list_container()
        container.pack_start(self.explanations, False, False, 5)

        # 词组短语标题
        self.phrases_label = __class__.label(
            '<span size="large" weight="bold" alpha="95%">词组短语：</span>')  # type: Gtk.Label
        container.pack_start(self.phrases_label, False, False, 5)

        # 词组短语列表
        self.phrases = __class__.list_container()
        container.pack_start(self.phrases, False, False, 0)

    def draw(self, query_result: QueryResult):
        # 查询内容
        if query_result.dict_link:
            self.query.set_markup(__class__.QUERY_TEMPLATE.format(query_result.dict_link, query_result.query))
        else:
            self.query.set_markup(__class__.QUERY_TEMPLATE_WITHOUT_LINK.format(query_result.query))

        # 基本释义
        if query_result.translation:
            self.translation.set_text(query_result.translation)
            self.translation.show()
        else:
            self.translation.hide()

        # 音标
        if query_result.phonetic:
            self.phonetic.set_markup('[{}]'.format(query_result.phonetic))
            self.phonetic.show()
        else:
            self.phonetic.set_markup('')
            self.phonetic.hide()

        # 基本翻译
        __class__.clear_list_container(self.explanations)
        if query_result.explanations:
            for item in query_result.explanations:
                label = __class__.auto_wrap_label(item)
                self.explanations.add(label)
            self.explanations_label.show()
            self.explanations.show_all()
        else:
            self.explanations_label.hide()
            self.explanations.hide()

        # 词组短语
        __class__.clear_list_container(self.phrases)
        if query_result.phrases:
            for item in query_result.phrases:
                if item['dict_link']:
                    markup = __class__.PHRASE_ITEM_TEMPLATE.format(item['dict_link'], item['key'], item['value'])
                else:
                    markup = __class__.PHRASE_ITEM_TEMPLATE_WITHOUT_LINK.format(item['key'], item['value'])
                label = __class__.auto_wrap_label(markup)
                self.phrases.add(label)
            self.phrases_label.show()
            self.phrases.show_all()
        else:
            self.phrases_label.hide()
            self.phrases.hide()

    # 基本 Label
    @staticmethod
    def label(markup: str = None) -> Gtk.Label:
        label = Gtk.Label(halign=Gtk.Align.START)
        if markup:
            label.set_markup(markup)
        return label

    # 单行 Label，空间不足以显示完整时，在末尾显示省略号
    @staticmethod
    def single_line_label() -> Gtk.Label:
        label = Gtk.Label(halign=Gtk.Align.START)
        label.set_single_line_mode(True)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        return label

    # 自动换行的 Label
    @staticmethod
    def auto_wrap_label(markup: str = None) -> Gtk.Label:
        label = Gtk.Label(halign=Gtk.Align.START)
        label.set_line_wrap(True)
        label.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        # 若不设置，换行时整个 Label 会左缩进
        label.set_xalign(0)
        if markup:
            label.set_markup(markup)
        label.show()
        return label

    # 列表容器
    @staticmethod
    def list_container() -> Gtk.Container:
        # Gtk.ListBox 样式有点问题（背景色不知如何取消），因此使用 Gtk.Box
        container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                            halign=Gtk.Align.START,
                            spacing=3)
        return container

    # 清空列表容器的内容
    @staticmethod
    def clear_list_container(container: Gtk.Container):
        for widget in container.get_children():
            widget.destroy()



import urllib.parse
from abc import ABC, abstractmethod
from typing import Optional, Type

from popupdict.config import *
from ..result import QueryResult


# Base class for query client
class AbstractQueryClient(ABC):
    # 客户端 ID，需唯一，用于区分客户端（配置文件、缓存等）
    id = 'abstract'
    # 配置类
    config_class = QueryClientConfiguration  # type: Type[QueryClientConfiguration]

    def __init__(self, config: QueryClientConfiguration):
        config.validate()
        self.config = config

    @abstractmethod
    def query(self, text: str) -> Optional[QueryResult]:
        pass

    # Escape URL path segment
    @staticmethod
    def escape_url_path(s: str):
        return urllib.parse.quote(s, safe='')



from collections import OrderedDict
from typing import Optional, Dict
import pickle
import os.path
import time

from .result import QueryResult
from popupdict.util import cache_dir, logger


# 缓存查询结果
# 不同查询客户端的缓存相互独立
# 使用内存做缓存，定期持久化到文件系统
# TODO 进程退出前自动持久化缓存
class QueryCache:
    # 单个客户端的最大缓存条目数
    MAX_COUNT = 1000
    # 将缓存持久化到硬盘的时间间隔，避免频繁写入硬盘。单位：秒
    PERSIST_INTERVAL = 30
    CACHE_FILE = os.path.join(cache_dir, 'query.cache')
    # 保存 QueryResult 的版本号，避免类结构变化
    CACHE_VERSION_FILE = os.path.join(cache_dir, 'query.cache-version')

    def __init__(self):
        self.clients = __class__.load_cache() or {}  # type: Dict[str, OrderedDict]
        self.last_persist_time = time.time()
        self.unpersisted_count = 0

        # QueryResult.VERSION 不会动态改变，因此只在程序启动时写入一次，不必每次持久化缓存时都重写
        with open(__class__.CACHE_VERSION_FILE, 'w') as f:
            f.write(str(QueryResult.VERSION))
        # 保证 cache 与 cache-version 一致
        self.persist()

    def get(self, client_id: str, query: str) -> Optional[QueryResult]:
        if client_id not in self.clients or query not in self.clients[client_id]:
            return
        self.clients[client_id].move_to_end(query)
        return self.clients[client_id][query]

    def put(self, client_id: str, query: str, query_result: QueryResult):
        self.unpersisted_count += 1
        if client_id not in self.clients:
            self.clients[client_id] = OrderedDict()
        self.clients[client_id][query] = query_result

        if len(self.clients[client_id]) > __class__.MAX_COUNT:
            self.clients[client_id].pop(False)

    def persist(self):
        # noinspection PyBroadException
        try:
            __class__.persist_cache(self.clients)
        except Exception:
            logger.exception('Query cache failed to persist')
        else:
            self.last_persist_time = time.time()
            self.unpersisted_count = 0
            logger.debug('Query cache persisted successfully')

    # 定时任务中调用
    def persist_if_necessary(self):
        if self.unpersisted_count > 0 and time.time() - self.last_persist_time > __class__.PERSIST_INTERVAL:
            self.persist()

    # 加载缓存
    @staticmethod
    def load_cache():
        if not os.path.exists(__class__.CACHE_VERSION_FILE) or not os.path.exists(__class__.CACHE_FILE):
            return
        # noinspection PyBroadException
        try:
            with open(__class__.CACHE_VERSION_FILE, 'r') as f:
                query_result_version = int(f.read().strip())
                if query_result_version != QueryResult.VERSION:
                    logger.debug('Ignore cache for version %d, current version is %d',
                                 query_result_version, QueryResult.VERSION)
                    return
            with open(__class__.CACHE_FILE, 'rb') as f:
                data = pickle.load(f)
                return data
        except Exception:
            return

    # 将缓存持久化到文件系统
    @staticmethod
    def persist_cache(obj):
        with open(__class__.CACHE_FILE, 'wb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

    # 清空缓存
    @staticmethod
    def clear_cache():
        if os.path.exists(__class__.CACHE_FILE):
            os.remove(__class__.CACHE_FILE)
        if os.path.exists(__class__.CACHE_VERSION_FILE):
            os.remove(__class__.CACHE_VERSION_FILE)



from setuptools import setup

import popupdict

setup(
    name="popupdict",
    version=popupdict.__version__,

    url='https://github.com/bianjp/popup-dict',
    author='Bian Jiaping',
    author_email='ssbianjp@gmail.com',
    license='MIT',
    keywords='dict youdao',
    description='Linux 下的划词翻译工具，支持有道智云等多种翻译服务',
    long_description='查看 `GitHub <https://github.com/bianjp/popup-dict>`_',

    packages=['popupdict'],
    include_package_data=True,
    zip_safe=True,

    platforms='Linux',
    python_requires='>= 3.6',
    install_requires=[
        'psutil',
        'requests',
    ],

    entry_points={
        'console_scripts': [
            'popup-dict = popupdict.main:main',
        ]
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: X11 Applications :: GTK',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: Chinese (Simplified)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Desktop Environment',
        'Topic :: Education',
    ],

)



import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')
from gi.repository import Gtk, Gdk, Gst, GLib, GObject, Pango

Gst.init()

__all__ = [
    'Gtk',
    'Gdk',
    'Gst',
    'GLib',
    'GObject',
    'Pango',
]



from configparser import SectionProxy

from ..exception import ConfigError


# 发音配置
class SpeechConfiguration:
    # noinspection PyUnresolvedReferences
    def __init__(self, section: SectionProxy):
        # 是否启用发音
        self.enabled = section.getboolean('enabled')
        # 是否自动播放发音
        self.auto_play = section.getboolean('auto_play')
        # 发音客户端 id
        self.client_id = section.get('client')

        if self.enabled and not self.client_id:
            raise ConfigError('No speech client specified')


# 各发音客户端配置基类
class SpeechClientConfiguration:
    def __init__(self, section: SectionProxy):
        # noinspection PyUnresolvedReferences
        self.request_timeout = section.getint('request_timeout')

    # Validate configuration, raise exception if invalid
    def validate(self):
        if not self.request_timeout or self.request_timeout <= 0:
            raise ConfigError('Invalid request_timeout value: {}'.format(repr(self.request_timeout)))

    def __repr__(self):
        result = self.__class__.__name__ + "("
        for k, v in self.__dict__.items():
            result += str(k) + '=' + str(v) + ', '
        result = result[:-2]
        result += ')'
        return result



from typing import Optional
import time
import os.path

from popupdict.config import *
from popupdict.util import logger
from .client import *
from .result import QueryResult
from .cache import QueryCache


# Query client adapter
class QueryAdapter:
    def __init__(self, config: Configuration):
        self.cache = QueryCache()
        for client_class in valid_clients:
            if client_class.id == config.query.client_id:
                config_section = config.clients.get(client_class.id, config.clients['default'])
                client_config = client_class.config_class(config_section)
                self.client = client_class(client_config)
                break

        if not self.client:
            raise ConfigError("Unknown query client: {}".format(repr(config.query.client_id)))

    def query(self, text: str) -> Optional[QueryResult]:
        start_time = time.time()
        logger.debug('[%s] Query started for query=%s', self.client.id, repr(text))

        # noinspection PyBroadException
        try:
            result = self.cache.get(self.client.id, text)
            if result:
                # TODO 完善路径的协议检查
                if result.speech_path and result.speech_path.startswith('/') and not os.path.exists(result.speech_path):
                    result.speech_path = None
                logger.debug('[%s] Cache hit for query=%s, result=%s', self.client.id, repr(text), repr(result))
                return result

            result = self.client.query(text)
            logger.debug('[%s] Query finished for query=%s, result=%s', self.client.id, repr(text), repr(result))
            if result:
                self.cache.put(self.client.id, text, result)
            return result
        except Exception:
            logger.exception('[%s] Query failed for query=%s', self.client.id, repr(text))
        finally:
            time_spent = (time.time() - start_time) * 1000
            logger.debug('[%s] Time spent: %fms', self.client.id, time_spent)

    def persist_cache(self):
        self.cache.persist_if_necessary()



import logging.handlers
import os.path

from popupdict.gtk import GLib

logger = logging.getLogger('popupdict')

formatter = logging.Formatter('[%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s', '%Y-%m-%d %H:%M:%S')

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

cache_dir = os.path.join(GLib.get_user_cache_dir(), 'popup-dict')
if not os.path.exists(cache_dir):
    os.mkdir(cache_dir, 0o755)
filename = os.path.join(cache_dir, 'popup-dict.log')
file_handler = logging.handlers.RotatingFileHandler(filename,
                                                    mode='a',
                                                    maxBytes=50 * 1024 * 1024,
                                                    backupCount=5,
                                                    encoding='utf-8')
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)
